/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <QtGui/QApplication>
#include <QtGui/QFileDialog>
#include <QtGui/QMessageBox>

#include <boost/tokenizer.hpp>

#include "wscout_gui_filter_dialog.hh"

namespace wscout
{

  namespace gui
  {

    FilterDialog::FilterDialog(QWidget* parent): Dialog (parent)
    {
      const QStyle&	s = *(qApp->style());
      QGridLayout&	l = layout();

      setupFilterRow(l);
      setupOutputFileNameRow(l, s);
      setupCommandRow(l);
      setupButtonsRow(l, s);

      process_.setProcessChannelMode(QProcess::MergedChannels);
      connect(&process_,
	      SIGNAL(finished(int, QProcess::ExitStatus)),
	      SLOT(processFinished(int, QProcess::ExitStatus)));
      connect(&process_,
	      SIGNAL(error(QProcess::ProcessError)),
	      SLOT(processError(QProcess::ProcessError)));
    }

    FilterDialog::~FilterDialog()
    {
      if (process_.state() == QProcess::Running)
	{
	  disconnect(&process_,	SIGNAL(finished(int, QProcess::ExitStatus)),
		     this, SLOT(processFinished(int, QProcess::ExitStatus)));
	  disconnect(&process_, SIGNAL(error(QProcess::ProcessError)),
		     this, SLOT(processError(QProcess::ProcessError)));
	  process_.kill();
	  process_.waitForFinished();
	}
    }

    QString
    FilterDialog::getFilteredTraceName(QWidget*		parent,
				       const QString&	input,
				       const QString&	title,
				       const QString&	default_out)
    {
      FilterDialog f (parent);

      f.setInputFileName(input);
      f.setWindowTitle(title);
      f.setOutputFileName(default_out);

      f.show();
      while (f.isVisible())
	QApplication::processEvents(QEventLoop::WaitForMoreEvents);

      if (f.result() == QDialog::Rejected)
	return QString ();

      return f.outputFileName();
    }

    void
    FilterDialog::okClicked()
    {
      typedef boost::escaped_list_separator<char>	separator;
      typedef boost::tokenizer<separator>		tokenizer;
      typedef tokenizer::const_iterator			iterator;

      const std::string	cmd = commandLine();
      const separator	sep ('\\', ' ', '"');
      const tokenizer	tok (cmd, sep);
      iterator		i = tok.begin();

      if (i == tok.end())
	{
	  QMessageBox::critical(this,
				tr("Invalid command"),
				tr("Please specify a non empty command."));
	  return;
	}

      const QString	arg0 = i->c_str();
      QStringList	args;

      while (++i != tok.end())
	{
	  if (not i->empty())
	    args << i->c_str();
	}

      setDefaultFilter(filter_.text());
      setDefaultCommand(command_.text());
      process_.start(arg0, args);

      ok_.setEnabled(false);
      running_.show();
    }

    void
    FilterDialog::cancelClicked()
    {
      done(Rejected);
    }

    void
    FilterDialog::processError(QProcess::ProcessError e)
    {
      bool	out = false;
      QString	msg = tr("<p><b>Error</b></p>");

      switch (e)
	{
	case QProcess::FailedToStart:
	  msg += tr("<p>The filter process failed to start. Either the"
		    " invoked program is missing, or you may have"
		    " insufficient permissions to invoke the program.</p>");
	  break;

	case QProcess::Crashed:
	  msg += tr("<p>The filter process crashed some time after starting"
		    " successfully.</p>");
	  out = true;
	  break;

	default:
	  msg += tr("<p>Unkown error running the filter process.</p>");
	  break;
	}
      msg += prettyCommandLine();
      if (out)
	msg += prettyProcessOutput();
      QMessageBox::critical(this, tr("Filter failed"), msg);
      running_.hide();
      ok_.setEnabled(true);
    }

    void
    FilterDialog::processFinished(int c, QProcess::ExitStatus s)
    {
      if (c != 0 or s == QProcess::CrashExit)
	{
	  const QString msg =
	    tr("<p><b>Error</b></p>"
	       "<p>The filter process reported failure.</p>") +
	    prettyProcessOutput() + prettyCommandLine();
	  QMessageBox::critical(this, tr("Filter failed"), msg);
	  running_.hide();
	  ok_.setEnabled(true);
	  return;
	}
      done(Accepted);
    }

    void
    FilterDialog::openOutputFileNameDialog()
    {
      QString s = QFileDialog::getSaveFileName(this, tr("Choose file"));

      if (not s.isNull())
	setOutputFileName(s);
    }

    void
    FilterDialog::setupFilterRow(QGridLayout& l)
    {
      filter_.setText(default_filter_);

      l.addWidget(makeLabel(tr("Filter")),	1, 1);
      l.addWidget(&filter_,			1, 2, 1, 3);
    }

    void
    FilterDialog::setupOutputFileNameRow(QGridLayout& l, const QStyle& s)
    {
      const QIcon&		i = s.standardPixmap(QStyle::SP_DirOpenIcon);
      QPushButton* const	b = new QPushButton (i, tr("C&hoose..."));

      b->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
      connect(b, SIGNAL(clicked(bool)), SLOT(openOutputFileNameDialog()));

      {
	const QFontMetrics	m (out_file_name_.font());

	out_file_name_.setMinimumWidth(m.width("m") * out_file_width);
      }

      l.addWidget(makeLabel(tr("Output file")),	2, 1);
      l.addWidget(&out_file_name_,			2, 2, 1, 2);
      l.addWidget(b,					2, 4);
    }

    void
    FilterDialog::setupCommandRow(QGridLayout& l)
    {
      command_.setText(default_command_);
      command_.setToolTip(tr("<p style=\"margin-bottom: 3px;"
			     "           margin-left: 8px;"
			     "           margin-right: 8px;\">"
			     "  <b>Available substitutions</b>"
			     "</p>"
			     "<p style=\"margin-top: 3px;"
			     "           margin-left: 8px;"
			     "           margin-right: 8px;\">"
			     "  %1 - input file<br />"
			     "  %2 - output file<br />"
			     "  %3 - filter"
			     "</p>"
			     "<p>"
			     "Separate arguments with spaces.  You may "
			     "escape spaces with backslash (\\) or pair of "
			     "double quotes (\")."
			     "</p>"));

      l.addWidget(makeLabel(tr("Command")),	3, 1);
      l.addWidget(&command_,			3, 2, 1, 3);
    }

    void
    FilterDialog::setupButtonsRow(QGridLayout& l, const QStyle& s)
    {
      // Running label.
      {
	running_.setText(tr("Running the filter process..."));
	running_.hide();
      }

      // Cancel.
      {
	const QIcon&	i = s.standardPixmap(QStyle::SP_DialogCancelButton);

	cancel_.setIcon(i);
	cancel_.setText(tr("&Cancel"));
	cancel_.setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
	connect(&cancel_, SIGNAL(clicked(bool)), SLOT(cancelClicked()));
      }

      // Ok.
      {
	const QIcon&	i = s.standardPixmap(QStyle::SP_DialogOkButton);

	ok_.setIcon(i);
	ok_.setText(tr("&Ok"));
	ok_.setDefault(true);
	connect(&ok_, SIGNAL(clicked(bool)), SLOT(okClicked()));
      }

      l.addWidget(&running_,	4, 2);
      l.addWidget(&cancel_,	4, 3);
      l.addWidget(&ok_,		4, 4);
    }

    QString
    FilterDialog::default_filter_;

    QString
    FilterDialog::default_command_ =
      tr("sh -c \"tshark -q -i- < '%1' -w '%2' -R '%3'\"");

  } // End of namespace gui.

} // End of namespace wscout.
