/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#ifndef WSCOUT_GUI_FILTER_DIALOG_HH_
# define WSCOUT_GUI_FILTER_DIALOG_HH_

# include <QtGui/QStyle>
# include <QtGui/QLabel>
# include <QtCore/QProcess>
# include <QtGui/QLineEdit>
# include <QtGui/QPushButton>

# include "wscout_gui_dialog.hh"

namespace wscout
{

  namespace gui
  {

    /// Dialog box to filter PCAP traces through external processes.
    class FilterDialog: public Dialog
    {
      Q_OBJECT

      /// Input PCAP file's name.
      Q_PROPERTY(QString inputFileName
		 READ inputFileName WRITE setInputFileName)

      /// Output PCAP file's name.
      Q_PROPERTY(QString outputFileName
		 READ outputFileName WRITE setOutputFileName)

      /// Length of the output file @c LineEdit widget (number of 'm' widths).
      enum { out_file_width = 25 };

    public:

      /// Default constructor.
      FilterDialog(QWidget* parent = 0);
      ~FilterDialog();

      /// @name Input file's name accessors
      //@{
      QString		inputFileName() const;
      void		setInputFileName(const QString&);
      //@}

      /// @name Output file's name accessors
      //@{
      QString		outputFileName() const;
      void		setOutputFileName(const QString&);
      //@}

      /// Set the default filter text for all FilterDialog instances.
      static void	setDefaultFilter(const QString&);

      /// Set the default command text for all FilterDialog instances.
      static void	setDefaultCommand(const QString&);

      /// Show a FilterDialog instance, filter a file, and retrieve
      /// the result's name
      static QString	getFilteredTraceName(QWidget* parent,
					     const QString& input,
					     const QString& title = QString (),
					     const QString& default_out =
					                      QString ());

    public slots:
      /// Click the @c OK button.
      void		okClicked();

      /// Click the @c Cancel button.
      void		cancelClicked();

    protected slots:
      /// Call when the filter process failed.
      void		processError(QProcess::ProcessError);

      /// Call when the filter process terminated.
      void		processFinished(int, QProcess::ExitStatus);

      /// Call when the user click on the file chooser.
      void		openOutputFileNameDialog();

    private:
      std::string	commandLine() const;
      QString		prettyCommandLine() const;
      QString		prettyProcessOutput();

      void		setupFilterRow(QGridLayout& l);
      void		setupOutputFileNameRow(QGridLayout& l, const QStyle& s);
      void		setupCommandRow(QGridLayout& l);
      void		setupButtonsRow(QGridLayout& l, const QStyle& s);

      QProcess		process_;
      QString		in_file_name_;
      QLineEdit		filter_;
      QLineEdit		out_file_name_;
      QLineEdit		command_;
      QPushButton	cancel_;
      QPushButton	ok_;
      QLabel		running_;

      static QString	default_filter_;
      static QString	default_command_;
    };

  } // End of namespace gui.

} // End of namespace wscout.

# include "wscout_gui_filter_dialog.hxx"

#endif // ! WSCOUT_GUI_FILTER_DIALOG_HH_
