/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#ifndef WSCOUT_GUI_PACKET_HH_
# define WSCOUT_GUI_PACKET_HH_

# include <vector>
# include <map>

# include <QtGui/QFrame>
# include <QtGui/QBoxLayout>

# include <wipal/pcap/descriptor.hh>

# include "wscout_gui_trace.hh"
# include "wscout_protocol_ids.hh"

namespace wscout
{

  namespace gui
  {

    // Forward declaration.
    struct AbstractHeader;

    /**
     * %Packet widget.
     *
     * Each instance represent a single packet. Protocol header
     * widgets register themselves to this class using the
     * registerProtocol() static method.
     *
     * @see AbstractHeader, HeaderBase
     */
    class Packet: public QFrame
    {
      Q_OBJECT

      enum
      {
	/// Space between header widgets.
	header_spacing		= 0,

	/// Packet's border width when selected.
	border_sel		= 4,

	/// Margin around headers when selected.
	packet_margin_sel	= 0,

	/// Packet's border width when unselected.
	border_unsel		= 1,

	/// Margin around headers when unselected.
	packet_margin_unsel	= 3

      };

    public:
      /// PCAP packet.
      typedef pcapxx::descriptor<>::iterator::value_type	packet;

      /// PCAP link type.
      typedef pcapxx::descriptor<>::link_type			link_type;

      /// Default constructor.
      Packet(QWidget* parent = 0);

      /// Setup the widget for a given link type, sequence number and
      /// PCAP packet.
      void		setup(const Trace& t, const packet& p);

      /// %Packet layout.
      QBoxLayout*	layout();

      /// Protocol header widget allocation function.
      typedef AbstractHeader* (*header_allocator)();

      /**
       * Register a %protocol header allocator for a given %protocol.
       *
       * All Packet instances share a common structure that maps
       * %protocol ids to %protocol header allocators.  Whenever
       * Packet::setup() encounters a registered %protocol, it calls the
       * corresponding allocator to create a %protocol header widget
       * to embed.
       *
       * This function registers %protocol ids and header allocators.
       *
       * @see AbstractHeader, header_allocator
       */
      static void	registerProtocol(protocol::id, header_allocator);

      /// Map from %protocol identifiers to %protocol header allocators.
      typedef std::map<protocol::id, header_allocator>	protocol_map;

      /**
       * Get the global %protocol map.
       *
       * Not a static variable to avoid cross-module initialization
       * order issues.
       */
      static protocol_map&	protocols();

    public slots:

      /// Select/unselect the packet.
      void			setSelected(bool state = true);

    private:
      int			lineWidth() const;
      int			margin() const;
      void			setupBorders();
      void			setupLayout();
      void			updateMinimumSize();

      std::vector<AbstractHeader*>	headers_;
      bool				selected_;
    };

  } // End of namespace gui.

} // End of namespace wscout.

# include "wscout_gui_packet.hxx"

#endif // ! WSCOUT_GUI_PACKET_HH_
