/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#ifndef WSCOUT_GUI_TRACE_HH_
# define WSCOUT_GUI_TRACE_HH_

# include <ostream>

# include <QtGui/QAbstractScrollArea>
# include <QtCore/QTimer>

# include <wipal/pcap/descriptor.hh>

namespace wscout
{

  namespace gui
  {

    // Forward declaration.
    struct Packet;

    /// Scrollable list of packets.
    class Trace: public QAbstractScrollArea
    {
      Q_OBJECT

      /// Position in the underlying @c pcapxx::descriptor instance.
      Q_PROPERTY(int position
		 READ position
		 WRITE setPosition
		 RESET setPosition
		 DESIGNABLE false
		 NOTIFY positionChanged(int)
		 USER true)

      Q_PROPERTY(bool swapContent
		 READ swapContent
		 WRITE setSwapContent
		 RESET setSwapContent
		 DESIGNABLE false)


      enum
      {
	/// Delay before updating after a call to setPosition().
	set_position_update_delay		= 0,

	/// Delay before updating after the scrollbar has been moved.
	scroll_contents_by_update_delay		= 20,

	/// Space between packets.
	packet_spacing				= 4,

	/// Vertical margins (top and bottom).
	trace_vmargin				= packet_spacing,

	/// Horizontal margins (left and right).
	trace_hmargin				= 8
      };

    public:
      /// Build a trace from a @c pcapxx::descriptor instance.
      explicit Trace(pcapxx::descriptor<>* trace = 0, QWidget* parent = 0);

      /// Copy constructor.
      Trace(const Trace& t);

      /// Destructor.
      virtual ~Trace();

      /// Change the underlying @c pcapxx::descriptor instance.
      void				setTrace(pcapxx::descriptor<>* t = 0);

      /// Retrieve the underlying @c pcapxx::descriptor instance.
      const pcapxx::descriptor<>*	trace() const;

      /// Retrieve the underlying @c pcapxx::descriptor instance.
      pcapxx::descriptor<>*		trace();

      /// Current position in the underlying @c pcapxx::descriptor instance.
      int		position() const;

      /// Last position in the underlying @c pcapxx::descriptor instance.
      int		lastPosition() const;

      /**
       * Get a summary of a given position as a string.
       *
       * This summary consists of the current position, the last
       * position, and the ratio of the two.
       *
       * @e e.g. @code 42 / 84 (50%) @endcode
       *
       * @param[in]	pos	The position to summarize.  By default the
       *			current position.
       */
      QString		positionString(int pos = -1) const;

      /**
       * Get whether this trace's packets contain swapped data.
       *
       * For instance, some traces hold Prism headers which endianness
       * is inconsistent with the PCAP endianness.  Thus the
       * swapContent property allows to override the PCAP endianness.
       */
      bool		swapContent() const;

      /// Get the last left clicked packet, or -1.
      int		lastLeftClicked() const;

    signals:
      /// The underlying @c pcapxx::descriptor instance changed.
      void		traceChanged();

      /// The current position changed.
      void		positionChanged(int);

      /// A packet has been left clicked.
      void		packetLeftClicked(int);

      /// A packet has been double left clicked.
      void		packetDoubleLeftClicked(int);

    public slots:
      /// Change the current position.
      void		setPosition(int p = 0);

      /**
       * Set the %swapContent property.
       *
       * @see swapContent().
       */
      void		setSwapContent(bool b = false);

      /// Update the widget's content.
      virtual void	update();

    protected:
      /// @name Qt overrided functions
      //@{
      virtual void	keyPressEvent(QKeyEvent*);
      virtual void	mouseMoveEvent(QMouseEvent*);
      virtual void	mousePressEvent(QMouseEvent*);
      virtual void	mouseDoubleClickEvent(QMouseEvent*);
      virtual void	resizeEvent(QResizeEvent*);
      virtual void	scrollContentsBy(int, int);
      //@}

    private:
      int		packetAt(const QPoint& pos) const;
      void		startUpdateTimer(int ms);
      void		resizeView(const QSize&	size,
				   bool		resize_pkts,
				   bool		update_if_needed);

      struct selection
      {
	/// Range of integer.
	struct range: public std::pair<int, int>
	{
	  /// Construct a range given its extremities.
	  range(int, int);
	};

	void		clear();
	bool		contains(int) const;
	std::ostream&	print(std::ostream&) const;
	selection&	toggle(int);

	selection&	operator = (int);
	selection&	operator = (const range&);

      private:
	typedef std::list<range>		list_of_ranges;
	typedef list_of_ranges::iterator	iterator;

	void		cleanup(iterator&);

	list_of_ranges	ranges_;
      };

      pcapxx::descriptor<>*	trace_;
      bool			swap_content_;
      QWidget			view_;
      int			pos_;
      QTimer			update_timer_;
      std::vector<Packet*>	pkts_;
      selection			selection_;
      int			last_selected_;
      int			last_left_clicked_;
    };

  } // End of namespace gui.

} // End of namespace wscout.

# include "wscout_gui_trace.hxx"

#endif // ! WSCOUT_GUI_TRACE_HH_
